#include "mainwindow.h"
#include "selectordialog.h"
#include "bookmark.h"
#include "addbookmarkdialog.h"
#include <QAction>
#include <QMenu>

#include <QtGui/QX11Info>
#include <X11/Xlib.h>
#include <X11/Xatom.h>


MainWindow::MainWindow(osso_context_t *context, QWidget *parent) : QMainWindow(parent)
{
    bookView = new BookView(this);
    setCentralWidget(bookView); 

    //Install eventFilter to bookView's viewport so we can intercept doubleclick event.
    bookView->viewport()->installEventFilter(this);

    
    connect(this, SIGNAL(volumeUpPressed()), bookView, SLOT(volumeUpPressed()));
    connect(this, SIGNAL(volumeDownPressed()), bookView, SLOT(volumeDownPressed()));

    toolbar = new QToolBar("toolbar", this);
    toolbar->setContextMenuPolicy(Qt::NoContextMenu);

    QAction *tocAction = toolbar->addAction("TOC");
    tocAction->setIcon(QIcon("/opt/mebook/data/toc_icon.png"));
    connect(tocAction, SIGNAL(triggered()), this, SLOT(tocNavigation()));

    QAction *bookmarkAction = toolbar->addAction("bookmarkAction");
    bookmarkAction->setIcon(QIcon("/opt/mebook/data/bm_icon.png"));
    connect(bookmarkAction, SIGNAL(triggered()), this, SLOT(viewBookmarks()));

    setContextMenuPolicy(Qt::CustomContextMenu);
    connect(this, SIGNAL(customContextMenuRequested(const QPoint&)), this, SLOT(openContextMenu(const QPoint&)));

    setWindowFlags(windowFlags() | Qt::Window);
    setAttribute(Qt::WA_Maemo5StackedWindow);
    addToolBar(toolbar);
    grabZoomKeys(true);

    osso_context = context;
    brightnessTimer.setInterval(60000);
    connect(&brightnessTimer, SIGNAL(timeout()), this, SLOT(forceBacklightOn()));
    brightnessTimer.start();
    setWindowTitle("MeBook");
}

MainWindow::~MainWindow()
{
    grabZoomKeys(false);
}

void MainWindow::openBook(const QString &filename)
{
    Book *book = new Book(filename);
    openBook(book);
}

void MainWindow::openBook(Book *book)
{
    if(bookView)
        bookView->setBook(book);
    currentBook = book;
    setWindowTitle(book->getTitle());
    forceBacklightOn();
}

void MainWindow::setFont(const QString &family, int size)
{
    if(bookView)
        bookView->setFont(family, size);
}

void MainWindow::setColorMode(int mode)
{
    if(bookView)
        bookView->setNightMode(mode);
}

void MainWindow::enableBacklightForcing(bool toggle)
{
    if(toggle)
        brightnessTimer.start();
    else
        brightnessTimer.stop();
}

void MainWindow::setViewMode(BookViewMode mode)
{
    bookView->setViewMode(mode);    
}

bool MainWindow::eventFilter(QObject *object, QEvent *event)
{
    if(event->type() == QEvent::MouseButtonDblClick)
    {
        if(!isFullScreen()){
            showFullScreen();
            toolbar->hide();            
        }
        else{
            showNormal();
            toolbar->show();
        }
    }else {
        return QObject::eventFilter(object, event);
    }
}

void MainWindow::keyPressEvent(QKeyEvent *event)
{
    if(event->key() == Qt::Key_F7)
        emit volumeUpPressed();
    else if(event->key() == Qt::Key_F8)
        emit volumeDownPressed();
    QMainWindow::keyPressEvent(event);
}

void MainWindow::tocNavigation()
{
    QList<QString> toc = currentBook->getTOC();
    //TODO: Show notofication banner if no TOC
    if(!toc.isEmpty()){
        SelectorDialog *dialog = new SelectorDialog(toc, this);
        dialog->setAttribute(Qt::WA_Maemo5StackedWindow);
        dialog->setWindowFlags(dialog->windowFlags() | Qt::Window);
        connect(dialog, SIGNAL(selectedString(QString)), bookView, SLOT(goToChapter(QString)));
        dialog->show();
    }


}

void MainWindow::viewBookmarks()
{
    QList<Bookmark> bm = currentBook->getBookmarks();
    if(bm.isEmpty())
        return;
    QList< QPair<QIcon, QString> > list;
    for(QList<Bookmark>::iterator iter = bm.begin(); iter != bm.end(); ++iter){
        Bookmark bookmark = *iter;
        QPixmap ip(32, 32); 
        QPainter painter(&ip);
        painter.setPen(Qt::NoPen);
        painter.setBrush(bookmark.getColor());
        painter.drawRect(ip.rect());
        QIcon icon(ip);

        QPair<QIcon, QString> pair;
        pair.first = icon;
        pair.second = bookmark.getName();

        list.push_back(pair);
    }
    SelectorDialog *dialog = new SelectorDialog(list, this);
    dialog->setAttribute(Qt::WA_Maemo5StackedWindow);
    dialog->setWindowFlags(dialog->windowFlags() | Qt::Window);
    connect(dialog, SIGNAL(selectedIndex(int)), bookView, SLOT(goToBookmark(int)));
    dialog->show();
}

void MainWindow::openContextMenu(const QPoint &pos)
{
    QMenu *menu = new QMenu();
    menu->addAction(tr("Add bookmark"), this, SLOT(addBookmark()));
    menu->exec(mapToGlobal(pos));
}

void MainWindow::addBookmark()
{
    AddBookmarkDialog *dialog = new AddBookmarkDialog(this); 
    if(dialog->exec() == QDialog::Accepted){
        bookView->addBookmark(dialog->getName(), dialog->getColor());        
    }
}


void MainWindow::grabZoomKeys(bool grab)
{
    if(!winId()){
        qDebug("Can't grab keys unless we have a window id");
        return;
    }

    unsigned long val = (grab) ? 1 : 0;
    Atom atom = XInternAtom(QX11Info::display(), "_HILDON_ZOOM_KEY_ATOM", False);
    if(!atom){
        qDebug("Unable to obtain _HILDON_ZOOM_KEY_ATOM.");
        return;
    }

    XChangeProperty(QX11Info::display(),
        winId(),
        atom,
        XA_INTEGER,
        32,
        PropModeReplace,
        reinterpret_cast<unsigned char*>(&val),
        1);

    grabKeyboard();
}

void MainWindow::closeEvent(QCloseEvent *event)
{
    bookView->saveProgress(); 
    QMainWindow::closeEvent(event);
}

void MainWindow::forceBacklightOn()
{
    qDebug() << "Forcing backlight on!";
    if(osso_context)
        osso_display_blanking_pause(osso_context);
}
